from typing import List, Any, Dict, Optional
from pydantic import BaseModel, validator
from enum import Enum
from xclientai.utils.logging import configure_logger
from xclientai.config import Config
from urllib.parse import urlparse


logger = configure_logger(__name__)


class Instance(str, Enum):
    G4DN_XLARGE = "g4dn.xlarge"


class Status(str, Enum):
    SUCCESSFUL = "successful"
    FAILED = "failed"
    RUNNING = "running"
    NOT_STARTED = "not_started"
    CANCELLED = "cancelled"
    DEPLOYING = "deploying"
    
    
class Credentials(BaseModel):
    aws_access_key_id: Optional[str]
    aws_secret_access_key: Optional[str]
    aws_region: Optional[str]
    cloud: str
    
    @validator("aws_access_key_id")
    def validate_aws_access_key_id(cls, v):
        return v
    
    @validator("aws_secret_access_key")
    def validate_aws_secret_access_key(cls, v):
        return v
    
    @validator("aws_region")
    def validate_aws_region(cls, v):
        return v
    
    @validator("cloud")
    def validate_cloud(cls, v, values, **kwargs):
        valid_clouds = ["aws", None]
        
        assert v in valid_clouds, f"Valid clouds are {valid_clouds}"
        
        if v == "aws":
            condition = "aws_access_key_id" in values and "aws_secret_access_key" in values and "aws_region" in values
            
            if not condition:
                logger.warning("AWS credentials have not been provided")
        
        return v


from typing import Dict, List, Optional, Union
from pydantic import BaseModel, validator
import re
from xclientai.dtypes.shared import Status, Credentials, Instance


class ContainerExecutionSpecs(BaseModel):
    image: str
    command: Optional[List[str]] = ["/bin/sh", "-c"]
    args: Optional[List[str]]
    env: Optional[Dict[str, str]]
    s3_code_dir: str

class PythonExecutionSpecs(BaseModel):
    requirements_file: Optional[str]
    main_py_file: str
    py_version: Optional[str] = "3.8"
    image: Optional[str]
    command: Optional[List[str]] = ["/bin/sh", "-c"]
    args: Optional[List[str]]
    env: Optional[Dict[str, str]]
    s3_code_dir: str
    
    @validator("py_version")
    def validate_py_version(cls, v):
        valid_py_versions = ["3.8"]
        
        if v is not None:
            assert v in valid_py_versions, "This Python version is not supported: {}. Supported Python versions: {}".format(v, valid_py_versions)
            
        return v
            
    @validator("image")
    def validate_image(cls, v, values):
        if values["py_version"] is not None:
            v = "python:3.8-slim-buster"
        else:
            raise ValueError("py_version cannot be None")
            
        return v    
    
    @validator("args")
    def validate_args(cls, v, values):
        if values["requirements_file"] is None:
            v = ["python {}".format(values["main_py_file"])]
        else:
            v = ["pip install -r {} ; python {}".format(
                values["requirements_file"],
                values["main_py_file"]
            )]
            
        return v
    
    class Config:
        validate_all = True
    
class ExecutionJob(BaseModel):
    job_name: str
    status: Status = Status.NOT_STARTED
    error: Optional[str]
    execution_spec: Union[ContainerExecutionSpecs, PythonExecutionSpecs]
    credentials: Credentials
    logs: Optional[List[str]]
    started: Optional[int]
    ended: Optional[int]
    instance: Optional[Instance] = Instance.G4DN_XLARGE
    num_instances: Optional[int] = 1
    
    
    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        
    @validator("job_name")
    def validate_job_name(cls, v):
        regex = "^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$"
        
        assert len(v) <= 30, "The job name cannot be longuer than 30 characters. Currently {}".format(len(v))
        
        assert bool(re.match(regex, v)), "The job_name must consist of lower case alphanumeric characters. Regex used for validation is '^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$'"
        
        return v  
    
    
    @validator("execution_spec")
    def validate_execution_spec(cls, v):
        if isinstance(v, PythonExecutionSpecs):
            v = ContainerExecutionSpecs(
                s3_code_dir=v.s3_code_dir,
                image=v.image,
                command=v.command,
                args=v.args,
                env=v.env
            )
            
        return v
    
    class Config:
        validate_all = True 



import requests
from requests.adapters import HTTPAdapter, Retry
import urllib3


urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def do_request(
    url, 
    http_method, 
    json_data=None, 
    headers=None,
    params=None
):
    session = requests.Session()
    retries = Retry(
        total=5,
        status_forcelist=[ 500, 502, 503, 504 ]
    )
    
    session.mount("http://", HTTPAdapter(max_retries=retries))
    
    params_fn = {
        "url": url, 
        "json": json_data,
        "timeout": 300,
        "headers": headers,
        "verify": False,
        "params": params
    }
    
    if http_method == "get":
        response = session.get(**params_fn)
    elif http_method == "post":
        response = session.post(**params_fn)
    elif http_method == "patch":
        response = session.patch(**params_fn)
    elif http_method == "delete":
        response = session.delete(**params_fn)
    elif http_method == "put":
        response = session.put(**params_fn)
    
    error = False
    try:
        response.raise_for_status()
    except Exception as ex:
        error = True
        
    if error:
        raise Exception(response.json()["message"])
    
    return response


from xclientai.utils.requests_utils import do_request
from xclientai.config import Config
from xclientai.dtypes.deployments import Deployment
from typing import List
import requests
from xclientai.dtypes.shared import Status
import time


class DeploymentsClient:
    
    @classmethod
    def get_deployments(cls) -> List[Deployment]:        
        response = do_request(
            url="{}/v1/deployments/".format(
                Config.DEPLOYMENTS_BASE_URL_X_BACKEND
            ),
            http_method="get"
        )
        
        list_deployment_dicts = response.json()["data"]
        deployments = []
        
        for deployment_dict in list_deployment_dicts:
            deployment = Deployment.parse_obj(deployment_dict)
            deployments.append(deployment)
            
        return deployments
    
    @classmethod
    def get_deployment_by_name(cls, deployment_name: str) -> Deployment:        
        response = do_request(
            url="{}/v1/deployments/{}".format(
                Config.DEPLOYMENTS_BASE_URL_X_BACKEND,
                deployment_name
            ),
            http_method="get"
        )
        
        deployment_dict = response.json()["data"]
        deployment = Deployment.parse_obj(deployment_dict)
        
        return deployment
    
    @classmethod
    def get_logs(cls, deployment_name: str) -> List[Deployment]:        
        response = do_request(
            url="{}/v1/deployments/{}/logs".format(
                Config.DEPLOYMENTS_BASE_URL_X_BACKEND,
                deployment_name
            ),
            http_method="get"
        )
        
        logs = response.json()["data"]
        
        return logs
    
    @classmethod
    def delete_deployment(cls, deployment_name: str) -> Deployment:
        response = do_request(
            url="{}/v1/deployments/{}".format(
                Config.DEPLOYMENTS_BASE_URL_X_BACKEND,
                deployment_name
            ),
            http_method="delete"
        )
        
        deployment_dict = response.json()["data"]
        deployment = Deployment.parse_obj(deployment_dict)
        return deployment
    
    @classmethod
    def create_deployment(cls, deployment: Deployment) -> Deployment:        
        deployment_dict = deployment.dict()
        
        response = do_request(
            url="{}/v1/deployments/".format(
                Config.DEPLOYMENTS_BASE_URL_X_BACKEND
            ),
            http_method="post",
            json_data=deployment_dict
        )
        
        returned_deployment_dict = response.json()["data"]
        returned_deployment = Deployment.parse_obj(returned_deployment_dict)
        
        return returned_deployment
       
    @classmethod
    def is_deployment_ready(cls, deployment_name: str):
        deployment = cls.get_deployment_by_name(deployment_name=deployment_name)
            
        if deployment.status == Status.RUNNING:
            response = requests.get(
                url=deployment.inference.is_ready_endpoint
            )
            
            if response.status_code == 200:
                is_deployment_ready = response.json().get("ready")
                
                if is_deployment_ready:
                    return True
                    
        return False
    
    @classmethod
    def wait_until_deployment_is_ready(cls, deployment_name: str, timeout=1000):
        sleep_time = 10
        total_time = 0
        while not cls.is_deployment_ready(deployment_name=deployment_name) and total_time < timeout:
            time.sleep(sleep_time)
            total_time += sleep_time


from xclientai.utils.requests_utils import do_request
from xclientai.config import Config
from xclientai.dtypes.optimization import AccelerationJob
from typing import List

class AccelerationJobsClient:
    
    @classmethod
    def get_jobs(cls, exclude_data_signature: bool = True) -> List[AccelerationJob]:        
        response = do_request(
            url="{}/v1/jobs/".format(
                Config.ACCELERATION_BASE_URL_X_BACKEND
            ),
            http_method="get",
            params={
                "exclude_data_signature": exclude_data_signature
            }
        )
        
        list_job_dicts = response.json()["data"]
        jobs = []
        
        for job_dict in list_job_dicts:
            job = AccelerationJob.parse_obj(job_dict)
            jobs.append(job)
            
        return jobs
    
    @classmethod
    def get_job_by_name(cls, job_name: str, exclude_data_signature: bool = True) -> AccelerationJob:        
        response = do_request(
            url="{}/v1/jobs/{}".format(
                Config.ACCELERATION_BASE_URL_X_BACKEND,
                job_name
            ),
            http_method="get",
            params={
                "exclude_data_signature": exclude_data_signature
            }
        )
        
        job_dict = response.json()["data"]
        job = AccelerationJob.parse_obj(job_dict)
        
        return job
    
    @classmethod
    def delete_job(cls, job_name: str, exclude_data_signature: bool = True) -> AccelerationJob:
        response = do_request(
            url="{}/v1/jobs/{}".format(
                Config.ACCELERATION_BASE_URL_X_BACKEND,
                job_name
            ),
            http_method="delete",
            params={
                "exclude_data_signature": exclude_data_signature
            }
        )
        
        job_dict = response.json()["data"]
        job = AccelerationJob.parse_obj(job_dict)
        return job
    
    @classmethod
    def create_job(cls, job: AccelerationJob, exclude_data_signature: bool = True) -> AccelerationJob:
        # Assert data in signature is not None
        for signature in job.acceleration.inputs + job.acceleration.outputs:
            assert signature.data is not None, "The data in signature cannot be None when creating a job"
        
        job_dict = job.dict()
        
        response = do_request(
            url="{}/v1/jobs/".format(
                Config.ACCELERATION_BASE_URL_X_BACKEND
            ),
            http_method="post",
            json_data=job_dict,
            params={
                "exclude_data_signature": exclude_data_signature
            }
        )
        
        returned_job_dict = response.json()["data"]
        returned_job = AccelerationJob.parse_obj(returned_job_dict)
        
        return returned_job
       


from typing import List, Any, Dict, Optional
from pydantic import BaseModel, Field, validator
import re
from xclientai.dtypes.shared import Status, Credentials, Instance
    
   
class ModelSpecs(BaseModel):
    model_path: str


class Batcher(BaseModel):
    max_batch_size: Optional[int]
    # Max latency in milliseconds
    max_lantecy: Optional[int] 
    timeout: Optional[int]


class Scaling(BaseModel):
    min: Optional[int] = 1
    max: Optional[int] = 1
    scale_metric: Optional[str]
    target_scaling_value: Optional[Any]
    
    
class DeploymentSpecs(BaseModel):
    image: Optional[str]
    batcher: Optional[Batcher]
    scaling: Optional[Scaling]
    
    
class Inference(BaseModel):
    base_endpoint: Optional[str]
    is_ready_endpoint: Optional[str]
    infer_endpoint: Optional[str]
    

class Deployment(BaseModel):
    deployment_name: str
    status: Status = Status.NOT_STARTED
    error: Optional[str]
    model_specs: ModelSpecs
    deployment_specs: Optional[DeploymentSpecs]
    inference: Optional[Inference]
    credentials: Credentials
    logs: Optional[str]
    started: Optional[int]
    ended: Optional[int]
    instance: Optional[Instance] = Instance.G4DN_XLARGE
    
    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        
    @validator("deployment_name")
    def validate_deployment_name(cls, v):
        regex = "^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$"
        
        assert len(v) <= 30, "The deployment name cannot be longuer than 30 characters. Currently {}".format(len(v))
        
        assert bool(re.match(regex, v)), "The deployment_name must consist of lower case alphanumeric characters. Regex used for validation is '^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$'"
        
        return v   



import logging


class CustomFormatter(logging.Formatter):
    """Logging colored formatter"""

    grey = '\x1b[38;21m'
    green = '\x1b[92m'
    blue = '\x1b[38;5;39m'
    yellow = '\x1b[38;5;226m'
    red = '\x1b[38;5;196m'
    bold_red = '\x1b[31;1m'
    reset = '\x1b[0m'

    def __init__(self, format):
        """Initializes the class

        Args:
            format: format of logger
        """
        super().__init__()
        self.fmt = format
        self.FORMATS = {
            logging.DEBUG: self.green + self.fmt + self.reset,
            logging.INFO: self.blue + self.fmt + self.reset,
            logging.WARNING: self.yellow + self.fmt + self.reset,
            logging.ERROR: self.red + self.fmt + self.reset,
            logging.CRITICAL: self.bold_red + self.fmt + self.reset
        }

    def format(self, record):
        """Given a record, format it

        Args:
            record: the record

        Returns:
            the formatted record
        """
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def configure_logger(name, fmt=None):
    # Create custom logger logging all five levels
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    # Define format for logs
    if fmt is None:
        fmt = '%(asctime)s | %(levelname)s | %(name)s %(lineno)d | %(message)s'

    # Create stdout handler for logging to the console (logs all five levels)
    stdout_handler = logging.StreamHandler()
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.setFormatter(CustomFormatter(fmt))

    # Add both handlers to the logger
    logger.addHandler(stdout_handler)

    return logger
